# imports


#  You can use this function to see how your model improves over time
def plot_history(history, metrics=None, ylim=None):
    plt.plot(history.history['loss'], label='training')
    plt.plot(history.history['val_loss'], label='testing')
    plt.title('Loss')
    if ylim:
        plt.ylim(ylim)
    plt.legend()
    plt.show()
    if metrics:
        for metric in metrics:
            plt.plot(history.history[metric], label=f'training {metric}')
            plt.plot(history.history[f'val_{metric}'], label=f'testing {metric}')
            plt.legend()
            plt.title(metric)
            if ylim:
                plt.ylim(ylim)
            plt.show()


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib
import scipy as sp
from mpl_toolkits.mplot3d import Axes3D
import numpy.linalg as la
import sys
sys.path.append("../../")
sys.path.append("..")



class cube:
    def __init__(self,
                 filename):
        self._cube = None
        self.filename = filename
        self.initialize()

    def initialize(self):
        atmDict = {1: 'H',
                   3: 'Li',
                   8: 'O',
                   11: 'Na',
                   55: 'Cs',
                   17: 'Cl',
                   35: 'Br',
                   53: 'I'}
        a = open(self.filename, "r")
        lin = a.readlines()
        beginningE = int(lin[2].split()[0]) + 6
        linNum = 1
        atmCt = 0
        self.atmStr = []
        ndim = []
        delta = []
        gridN = 0
        for line in lin:
            if linNum == 3:
                splt = line.split()
                info = [float(q) for q in splt]
                nAtoms = int(info[0])
                self.atmAr = np.zeros((nAtoms, 3))
                # atmStr = np.tile(["O","H","H"],nAtoms//3)
                origin = np.array(info[1:-1])
            elif 4 <= linNum <= 6:
                splt = line.split()
                xyzZ = [float(q) for q in splt]
                ndim.append(int(xyzZ[0]))
                if linNum == 4:
                    delta = xyzZ[1]
            elif 7 <= linNum <= beginningE:
                splt = line.split()
                xyzZ = [float(q) for q in splt][1:]
                self.atmStr.append(atmDict[int(splt[0])])
                self.atmAr[atmCt] = xyzZ[1:]
                atmCt += 1
            elif linNum > beginningE:
                pass
            linNum += 1
        a.close()
        k = pd.read_table(self.filename, delim_whitespace=True, header=None, skiprows=np.arange(beginningE))
        xx = k.to_numpy()
        xxp = xx[np.logical_not(np.isnan(xx))]
        self.density = np.reshape(xxp, ndim)
        self.cdsX = np.linspace(origin[0], origin[0] + delta * ndim[0], num=ndim[0])
        self.cdsY = np.linspace(origin[1], origin[1] + delta * ndim[1], num=ndim[1])
        self.cdsZ = np.linspace(origin[2], origin[2] + delta * ndim[2], num=ndim[2])
        self.cds = {'x' : self.cdsX, 'y': self.cdsY, 'z':self.cdsZ}
        self.arr = np.meshgrid(self.cdsX, self.cdsY, self.cdsZ, indexing='ij')
        dxO = self.cdsX[1] - self.cdsX[0]
        dyO = self.cdsY[1] - self.cdsY[0]
        dzO = self.cdsZ[1] - self.cdsZ[0]
        print("DXDYDZO", dxO, dyO, dzO, dxO * dyO * dzO)
        nbins = 40
        shiftL = 1.0
        shiftR = 1.0
        cylRad = 1.0




import sys
sys.path.append("/Users/coire/McCoy/") #Path to McUtils, References..etc
import numpy as np
import copy
import McUtils.GaussianInterface as GI
import os
import matplotlib.pyplot as plt
from Chem_Tools import misc, density_cube as dc


class GaussianResults:
    '''
    The purpose of this class is to combine the results from fchk and log files into a single object
    What gets pulled from where depends on the type of job run
    I am assuming this class will be updated a lot moving forward
    '''
    def __init__(self, files, **kwargs):
        self.params = kwargs
        if len(files) == 0:
            raise Exception('Nothing to interpret.')
        self.files = files
        if not isinstance(files, list): self.files = [self.files]
        self._MP2Energy = None
        self._input_zmat = None
        self.force_constants = None

    @property
    def MP2Energy(self):
        if self._MP2Energy is None:
            self._MP2Energy = self.get_MP2energy()
        return self._MP2Energy

    @property
    def input_zmat(self):
        if self._input_zmat is None:
            self._input_zmat = self.get_input_zmat()
        return self._input_zmat

    def get_MP2energy(self):
        energy = []
        for file in self.files:
            with GI.GaussianFChkReader(file + '.fchk') as reader:
                parse = reader.parse("MP2 Energy")
            ens = parse["MP2 Energy"]
            energy.append(ens)
        e = np.array(energy)
        return e

    def get_input_zmat(self):
        zmats = []
        for f in self.files:
            with GI.GaussianLogReader(f + '.log') as reader:
                parse = reader.parse("InputZMatrix")
            if parse["InputZMatrix"] is None:
                print(f'could not pull the z-matrix from {f}')
                print(parse["InputZMatrix"])
            new_zmat = misc.zmat(g_str=parse["InputZMatrix"])
            zmats.append(new_zmat)
        return zmats

    def pull_forces(self):
        forces = []
        for f in self.files:
            with GI.GaussianLogReader(f + '.log') as reader:
                parse = reader.parse("ForceConstants")
            forces.append(parse["ForceConstants"])
        self.force_constants=forces



class GaussianJob:
    def __init__(self, params, filename):
        self.params = params
        self.filename=filename

    def save_job(self):
        with open(self.filename, 'w', newline='\n') as f:
            for key in self.params:
                f.write(self.params[key])






class GLogInterpreter:
    def __init__(self, log_files):
        self.log_files = log_files
        if not isinstance(log_files, list): self.log_files = [self.log_files]
        self._Energy = None
        self._intCoords = None
        self._dipole_moms = None
        self._input_zmat = None
        self.force_constants = None
        self.NBO_Summary = None

    @property
    def Energy(self):
        if self._Energy is None:
            self._Energy = self.get_Ens()
        return self._Energy

    @property
    def intCoords(self):
        if self._intCoords is None:
            self._intCoords = self.get_intCoords()
        return self._intCoords

    @property
    def dipole_moms(self):
        if self._dipole_moms is None:
            self._dipole_moms = self.get_dipole_moms()
        return self._dipole_moms

    @property
    def input_zmat(self):
        if self._input_zmat is None:
            self._input_zmat = self.get_input_zmat()
        return self._input_zmat

    def get_Ens(self):
        energy_array = []
        for f in self.log_files:
            with GI.GaussianLogReader(f) as reader:
                parse = reader.parse("OptimizedScanEnergies")
            temp_energy_array, temp_coords = parse["OptimizedScanEnergies"]
            if np.any(energy_array):
                energy_array = np.append(energy_array, temp_energy_array)
            else:
                energy_array = copy.deepcopy(temp_energy_array)
        return energy_array

    def get_intCoords(self):
        coords = {}
        for f in self.log_files:
            with GI.GaussianLogReader(f) as reader:
                parse = reader.parse("OptimizedScanEnergies")
            temp_energy_array, temp_coords = parse["OptimizedScanEnergies"]
            if coords:
                for key in coords.keys():
                    coords[key] = np.append(coords[key],temp_coords[key])
            else:
                coords = copy.deepcopy(temp_coords)
        return coords

    def get_dipole_moms(self):
        dipoles = {}
        for f in self.log_files:
            with GI.GaussianLogReader(f) as reader:
                parse = reader.parse("DipoleMoments")
            #print(parse["DipoleMoments"])
            temp_dipoles = parse["DipoleMoments"]

            if dipoles:
                for key in dipoles.keys():
                    dipoles[key] = np.append(dipoles[key],temp_dipoles[key])
            else:
                dipoles = copy.deepcopy(temp_dipoles)
        return dipoles

    def get_input_zmat(self):
        zmats = []
        for f in self.log_files:
            with GI.GaussianLogReader(f) as reader:
                parse = reader.parse("InputZMatrix")
            zmats.append(misc.zmat(g_str= parse["InputZMatrix"]))
        return zmats

    def pull_forces(self):
        for f in self.log_files:
            with GI.GaussianLogReader(f) as reader:
                parse = reader.parse("ForceConstants")
            forces= parse["ForceConstants"]
        self.force_constants=forces

    def pull_NBO(self):
        d = {}
        ex = {'occupancy': [],
                 'file': []}
        for f in self.log_files:
            with GI.GaussianLogReader(f) as reader:
                parse = reader.parse("NBOSummary")
            NBO= parse["NBOSummary"]
            for o in range(len(NBO[0])):
                if NBO[0][o] in d:
                    d[NBO[0][o]]['occupancy'].append(NBO[1][o])
                    d[NBO[0][o]]['file'].append(f)
                else:
                    new_d = copy.deepcopy(ex)
                    new_d['occupancy'].append(NBO[1][o])
                    new_d['file'].append(f)
                    d[NBO[0][o]] = new_d
        self.NBO_Summary = d





class GLogPlotter:
    def __init__(self, files):
        self.files = files
        self._data = None

    @property
    def data(self):
        if self._data is None:
            self._data = self.get_data()
        return self._data

    def get_data(self):
        data = GLogInterpreter(self.files)
        return data

    def plot_energies(self):
        x = self.data.Energy
        fig = plt.figure()
        plt.plot(x, '.')

    def plot_energy_v_coord(self, xCoord):
        y = self.data.Energy
        x = self.data.intCoords[xCoord]
        plt.plot(x, y, ".")

    def plot_energy_v_twoCoords(self, xCoord, yCoord):
        x = self.data.intCoords[xCoord]
        y = self.data.intCoords[yCoord]
        z = self.data.Energy
        fig = plt.figure()
        ax = plt.axes(projection='3d')
        ax.scatter3D(x, y, z, 'gray')
        ax.set_xlabel(xCoord)
        ax.set_ylabel(yCoord)
        ax.set_zlabel('Energy')

    def plot_coord_v_coord(self, xCoord, yCoord):
        x = self.data.intCoords[xCoord]
        y = self.data.intCoords[yCoord]
        fig = plt.figure()
        ax = plt.axes()
        ax.plot(x,y,'.')
        ax.set_xlabel(xCoord)
        ax.set_ylabel(yCoord)

class FchkInterpreter:
    # Copied directly from https://github.com/rmhuch/QOOH/blob/master/FChkInterpreter.py
    #then modified
    def __init__(self, fchks, cubes=False,  **kwargs):
        self.params = kwargs
        if len(fchks) == 0:
            raise Exception('Nothing to interpret.')
        self.fchks = fchks
        if not isinstance(fchks, list): self.fchks = [self.fchks]
        self._hessian = None
        self._cartesians = None  # dictionary of cartesian coordinates keyed by (x, y) distances
        self._gradient = None
        self._MP2Energy = None
        self._atomicmasses = None
        self._atomicNumbers = None
        self._mullikens = None
        self.cubes= cubes
        if self.cubes:
            self.initializeCubes()

    @property
    def cartesians(self):
        if self._cartesians is None:
            self._cartesians = self.get_coords()
        return self._cartesians

    @property
    def hessian(self):
        if self._hessian is None:
            self._hessian = self.get_hess()
        return self._hessian

    @property
    def gradient(self):
        if self._gradient is None:
            self._gradient = self.get_grad()
        return self._gradient

    @property
    def MP2Energy(self):
        if self._MP2Energy is None:
            self._MP2Energy = self.get_MP2energy()
        return self._MP2Energy

    @property
    def atomicmasses(self):
        if self._atomicmasses is None:
            self._atomicmasses = self.get_mass()
        return self._atomicmasses

    @property
    def atomicNumbers(self):
        if self._atomicNumbers is None:
            self._atomicNumbers = self.get_atomicNumbers()
        return self._atomicNumbers

    @property
    def mullikens(self):
        if self._mullikens is None:
            self._mullikens = self.get_mullikens()
        return self._mullikens

    def get_coords(self):
        """Uses McUtils parser to pull cartesian coordinates
            :returns coords: nx3 coordinate matrix"""
        crds = []
        for fchk in self.fchks:
            with GI.GaussianFChkReader(fchk) as reader:
                parse = reader.parse("Coordinates")
            coords = parse["Coordinates"]
            crds.append(coords)
        c = np.array(crds)
        if c.shape[0] == 1:
            c = np.squeeze(c)
        return c

    def get_hess(self):
        """Pulls the Hessian (Force Constants) from a Gaussian Frequency output file
            :arg chk_file: a Gaussian Frequency formatted checkpoint file
            :returns hess: full Hessian of system as an np.array"""
        hess = []
        for fchk in self.fchks:
            with GI.GaussianFChkReader(fchk) as reader:
                parse = reader.parse("ForceConstants")
            forcies = parse["ForceConstants"]
            hess.append(forcies)
        h = np.array(hess)
        # ArrayPlot(forcies.array, colorbar=True).show()
        return h

    def get_grad(self):
        for fchk in self.fchks:
            with GI.GaussianFChkReader(fchk) as reader:
                parse = reader.parse("Gradient")
            grad = parse["Gradient"]
        return grad

    def get_MP2energy(self):
        energy = []
        for fchk in self.fchks:
            with GI.GaussianFChkReader(fchk) as reader:
                parse = reader.parse("MP2 Energy")
            ens = parse["MP2 Energy"]
            energy.append(ens)
        e = np.array(energy)
        return e

    def get_mass(self):
        for fchk in self.fchks:
            with GI.GaussianFChkReader(fchk) as reader:
                parse = reader.parse("AtomicMasses")
            mass_array = parse["AtomicMasses"]
        return mass_array

    def get_atomicNumbers(self):
        for fchk in self.fchks:
            with GI.GaussianFChkReader(fchk) as reader:
                parse = reader.parse("AtomicNumbers")
            nums = parse["AtomicNumbers"]
        return nums

    def get_mullikens(self):
        mulls = []
        for fchk in self.fchks:
            with GI.GaussianFChkReader(fchk) as reader:
                parse = reader.parse("Mulliken Charges")
            mulls.append(parse["Mulliken Charges"])
        return np.array(mulls)

    def initializeCubes(self):
        ed_cds = []
        ed_vals = []
        for f in self.fchks:
            print(f)
            cube = dc.cube(os.path.splitext(f)[0] + '.cube')
            ed_cds.append(cube.cds)
            ed_vals.append(cube.density)
        self.ed_cds = np.array(ed_cds)
        self.ed_vals = np.array(ed_vals)



#class NormalModes:
    #def __init__(self):

#class ZMatrix:

class CoirePlots:
    def __init__(self, *data, type=None):
        self.data = data
        self.type = type
        self._axSettings = None
        self._pltSettings = None

    @property
    def axSettings(self):
        if self._axSettings is None:
            self._axSettings = self.emptyAxSettings()
        return self._axSettings

    @property
    def pltSettings(self):
        if self._pltSettings is None:
            self._pltSettings = self.emptyPltSettings()
        return self._pltSettings

    def emptyAxSettings(self):
        settings = {'xlabel': None,
                    'ylabel': None,
                    'title': None}
        return settings

    def plotType(self, type):
        self.type = type

    #def setSettings(self, ax):

    def plot(self, **kwargs):
        if self.type == 'scatter2d':
            self.fig = plt.figure()
            self.ax = plt.axes()
            self.ax.scatter(self.data[0], self.data[1], kwargs)
            if self.axSettings is not None:
                self.setAxSettings()
            if self.pltSettings is not None:
                self.setPltSettings()

    def setAxSettings(self):
        self.ax.set_xlabel(self.axSettings['xlabel'])
        self.ax.set_ylabel(self.axSettings['xlabel'])
        self.ax.title(self.axSettings['title'])





import pickle

def pickle_save(dObj, sFilename):
    '''Given an object and a file name, write the object to the file using pickle.'''

    sFilename = sFilename + ".pickle"
    f = open(sFilename, "wb")
    p = pickle.Pickler(f)
    p.dump(dObj)
    f.close()

def pickle_load(sFilename):
    '''Given a file name, load and return the object stored in the file.'''

    sFilename = sFilename + ".pickle"
    f = open(sFilename, "rb")
    u = pickle.Unpickler(f)
    dObj = u.load()
    f.close()
    return dObj

def save(object, method, filename):
    if method == 'pickle':
        pickle_save(dObj=object, sFilename=filename)



def load(method, filename):
    if method == 'pickle':
        obj = pickle_load(sFilename=filename)
    if method == 'npz':
        obj = np.load(filename)
    return obj


import glob
import sys
from Chem_Tools import misc
import os
sys.path.append('../../mccoy_hoono/')
import numpy as np

if __name__ == "__main__":
    atoms = ['H', 'O', 'O', 'N', 'O']
    specs = np.array([[1, 2, 3, 4], [1, 2, 3], [1, 2]])
    names = np.array([['r1', 'r2', 'r3', 'r4'], ['a1', 'a2', 'a3'], ['d1', 'd2']])
    vals = np.array([[0.98537664, 1.42673666, 1.38110764, 1.19478846],
                     [100.6355057, 113.13777052, 114.33392022],
                     [0., 0.]])
    units = ["Angstroms", "Degrees"]
    hoono_eq = misc.zmat(atoms=atoms, easy_build_dict={'specs':specs, 'names':names, 'vals':vals, 'units':units})
    # hoono_eq_log = gt.GLogInterpreter(log_files='../mccoy_hoono/hoono_eq.log')
    # hoono_eq_log.pull_forces()
    # params = {
    #                 'header': f'%nproc=28\n%mem=120GB\n%chk=',
    #                 'job': '#p SP mp2/aug-cc-pvtz density=current\n\n',
    #                 'description': 'title\n\n',
    #                 'charge': '0 1\n',
    #                 'molecule': '',
    #                 'variables': 'Variables\n',
    #                 'vars': '',
    #                 'whitespace': '\n'*10}
    # disps = misc.target_displacements(eq_zmat=hoono_eq,
    #                           force_constants=hoono_eq_log.force_constants)
    # misc.write_displaced_zmats(eq_zmat=hoono_eq,
    #                            force_constants=hoono_eq_log.force_constants,
    #                            fname_base='hoono',
    #                            params=params,
    #                            job_file_path='../mccoy_hoono/explore_disps/')

    # hoono_fchk = gt.FchkInterpreter(fchks=['../mccoy_hoono/explore_disps/hoono_a1_95p93353.fchk'])
    # hoono_log = gt.GLogInterpreter(log_files=['../mccoy_hoono/explore_disps/hoono_a1_95p93353.log'])
    # with GI.GaussianLogReader('../mccoy_hoono/hoono_eq.log') as reader:
    #     parse = reader.parse("OptimizedScanEnergies")
    # z = misc.zmat(g_str='\n Charge =  0 Multiplicity = 1\n H\n O                    1    r1\n O                    2    r2       1    a1\n N                    3    r3       2    a2       1    d1       0\n O                    4    r4       3    a3       2    d2       0\n       Variables:\n  r1                    0.96898                  \n  r2                    1.45492                  \n  r3                    1.45878                  \n  r4                    1.17073                  \n  a1                   95.93353                  \n  a2                  102.29752                  \n  a3                  109.06955                  \n  d1                  180.                       \n  d2                  180.')
    #misc.zmat()
    # result = gt.GaussianResults(files=['../mccoy_hoono/explore_disps/hoono_d2_176p1845'])
    # result = gt.GaussianResults(files=[os.path.splitext(val)[0] for val in glob.glob("../mccoy_hoono/explore_disps/*.gjf")])
    #
    hoono_eq_energy = -280.4467347
    results = misc.pull_displaced_results(eq_zmat=hoono_eq,
                                          eq_energy=hoono_eq_energy,
                                          files=[os.path.splitext(val)[0] for val in glob.glob("../mccoy_hoono/single_coord_freq/hoono_*_*.log")],
                                          eq_log = '../mccoy_hoono/hoono_eq')
    d = misc.fourth_ders(results)
    # result = gt.GaussianResults(files=files)
    # result = gt.GaussianResults(files=[os.path.splitext(val)[0] for val in glob.glob("../mccoy_hoono/30cm_disp/*.gjf")])
    # with GI.GaussianLogReader('../hocl/oop_s_rOH_opt/clhocl/clhocl_0_sp.log') as reader:
    #     parse = reader.parse("NBOSummary")
    # l = ['../hocl/oop_s_rOH_opt/clhocl/clhocl_0_sp.log',
    #      '../hocl/oop_s_rOH_opt/clhocl/clhocl_5_sp.log',
    #      '../hocl/oop_s_rOH_opt/clhocl/clhocl_10_sp.log',
    #      '../hocl/oop_s_rOH_opt/clhocl/clhocl_15_sp.log',
    #      '../hocl/oop_s_rOH_opt/clhocl/clhocl_20_sp.log',
    #      '../hocl/oop_s_rOH_opt/clhocl/clhocl_25_sp.log',
    #      '../hocl/oop_s_rOH_opt/clhocl/clhocl_30_sp.log',
    #      '../hocl/oop_s_rOH_opt/clhocl/clhocl_35_sp.log',
    #      '../hocl/oop_s_rOH_opt/clhocl/clhocl_40_sp.log',
    #      '../hocl/oop_s_rOH_opt/clhocl/clhocl_45_sp.log',
    #      '../hocl/oop_s_rOH_opt/clhocl/clhocl_50_sp.log']
    # N = gt.GLogInterpreter(log_files=l)
    # N.pull_NBO()


    print("Hello")



# Imports
from ipywidgets import interactive
import seaborn as sns
import ipywidgets as widgets
import matplotlib.pyplot as plt
import math
import pandas as pd
import numpy as np


class missing_vals:
    def __init__(self, df):
        self.df = df
        self.isna_df = self.find_isna_cols()
        self.make_widget()

    def find_isna_cols(self):
        isna_d = {'Feature': [],
                  'dtype': [],
                  'Missing Vals': [],
                  'Percentage': []}
        for key in self.df:
            if self.df[key].isna().sum() > 0:
                isna_d['Feature'].append(key)
                isna_d['dtype'].append(self.df[key].dtype)
                isna_d['Missing Vals'].append(self.df[key].isna().sum())
                isna_d['Percentage'].append(self.df[key].isna().sum() / len(self.df))
        return pd.DataFrame(isna_d)

    def list_isna(self):
        display(self.isna_df)

    def unique(self, column):
        print(self.isna_df.loc[self.isna_df['Feature'] == column])
        print()
        print(self.df[column].unique())

    def make_widget(self):
        dd = widgets.Dropdown(options=self.isna_df['Feature'].unique(),
                              value=self.isna_df['Feature'].unique()[0],
                              description="Feature")
        self.w = interactive(self.unique, column=dd)

class Countplot:
    def __init__(self, df):
        self.df = df
        self.make_widget()

    def draw_countplot(self, column, hue, check_hue):
        sns.set(font_scale=1.2)
        if check_hue == False:
            p = sns.countplot(data=self.df, x=column)
            if len(self.df[column].unique()) > 4:
                p.tick_params(axis='x', rotation=60)
        else:
            p = sns.countplot(data=self.df, x=column, hue=hue)
            if len(self.df[column].unique()) > 4:
                p.tick_params(axis='x', rotation=60)
        plt.figure(figsize=(20, 20))

    def make_widget(self):
        cat_columns = [column for column in self.df.columns if self.df[column].dtype == 'object']
        dd = widgets.Dropdown(options=cat_columns,
                              value=cat_columns[0],
                              description="Feature")
        dd_hue = widgets.Dropdown(options=cat_columns,
                              value=cat_columns[0],
                              description="Hue")
        check_hue = widgets.Checkbox(False, description='By Hue?')

        self.w = interactive(self.draw_countplot, column=dd, hue=dd_hue, check_hue=check_hue)

class simple_box:
    def __init__(self, df):
        self.df = df
        self.make_widget()

    def draw_bar(self, column):
        sns.set(font_scale=1.2)
        p = sns.boxplot(data=self.df, y = column)
        plt.figure(figsize=(20, 20))

    def make_widget(self):
        num_columns = [column for column in self.df.columns if ((self.df[column].dtype == 'int64')|(self.df[column].dtype == 'float'))]
        dd = widgets.Dropdown(options=num_columns,
                              value=num_columns[0],
                              description="Feature")
        """dd_hue = widgets.Dropdown(options=num_columns,
                                  value=num_columns[0],
                                  description="Hue")
        check_hue = widgets.Checkbox(False, description='By Hue?')"""

        self.w = interactive(self.draw_bar, column=dd)


# import math
# import pandas as df
# import numpy as np
# https://napsterinblue.github.io/notes/python/viz/subplots/
# https://www.python-graph-gallery.com/scatterplot-with-regression-fit-in-matplotlib
def plot_linearity(df, y):
    num_df = df.select_dtypes(include='number').drop(columns=y)
    n_cols = len(num_df.columns)

    fig, axes = plt.subplots(math.ceil(n_cols / 4), 4, figsize=(16, 8))
    for idx, (col, ax) in enumerate(zip(num_df.columns, axes.flatten())):
        ax.scatter(num_df[col], df[y])
        ax.set_title(col)
        b, a = np.polyfit(num_df[col], df[y], deg=1)
        fit_x = np.linspace(num_df[col].min(), num_df[col].max(), 100)
        ax.plot(fit_x, a + b * fit_x, color="k", lw=2.5);
        plt.subplots_adjust(wspace=.5, hspace=.5)

    else:
        [ax.set_visible(False) for ax in axes.flatten()[idx + 1:]]



import sys
sys.path.append("/Users/coire/McCoy/") #Path to McUtils, References..etc
import numpy as np
import copy
from McUtils import Data as McData
from Chem_Tools import gaussian_tools as gt


def fourth_ders(results):
    der = {'disp_vals': [],
           'fourth_ders': []}

    derivatives = {}

    for q in results:
        new_q = copy.deepcopy(der)
        eq_index = int((len(results[q]['disp_vals']) - 1) / 2)
        for i in range(1, eq_index + 1):
            h = results[q]['disp_vals'][eq_index + i]
            new_q['disp_vals'].append(h)
            new_q['fourth_ders'].append((results[q]['force_constants'][eq_index + i] +results[q]['force_constants'][eq_index - i] -2 * results[q]['force_constants'][eq_index]) /h**2)
        derivatives[q] = new_q
    return derivatives


def pull_displaced_results(eq_zmat, eq_energy, files, eq_log):
    '''
    eq_energy in hartree
    outputs displaced energy in wavenumbers'''
    r = {'filenames': [],
         'disp_vals': [],
         'disp_units': [],
         'disp_energies': [],
         'force_constants': []
         }
    results = {}
    gr = gt.GaussianResults(files = files)
    gr.pull_forces()
    #results['filenames'] = files
    #results['disp_energies'] = gr.MP2Energy
    for i in range(len(files)):
        #find displaced coord
        for e in range(len(eq_zmat.coords)):
            if round(gr.input_zmat[i].coords[e].val,5) != round(eq_zmat.coords[e].val,5):
                if gr.input_zmat[i].coords[e].name in results:
                    results[gr.input_zmat[i].coords[e].name]['filenames'].append(files[i])
                    results[gr.input_zmat[i].coords[e].name]['disp_vals'].append(gr.input_zmat[i].coords[e].val - eq_zmat.coords[e].val)
                    results[gr.input_zmat[i].coords[e].name]['disp_units'].append(gr.input_zmat[i].coords[e].units)
                    results[gr.input_zmat[i].coords[e].name]['disp_energies'].append((gr.MP2Energy[i]-eq_energy)*McData.UnitsData.convert( "Hartrees","Wavenumbers"))
                    results[gr.input_zmat[i].coords[e].name]['force_constants'].append(gr.force_constants[i][e][e])
                else:
                    new_result = copy.deepcopy(r)
                    new_result['filenames'].append(files[i])
                    new_result['disp_vals'].append(gr.input_zmat[i].coords[e].val - eq_zmat.coords[e].val)
                    new_result['disp_units'].append(gr.input_zmat[i].coords[e].units)
                    new_result['disp_energies'].append((gr.MP2Energy[i]-eq_energy)*McData.UnitsData.convert( "Hartrees","Wavenumbers"))
                    new_result['force_constants'].append(gr.force_constants[i][e][e])
                    results[gr.input_zmat[i].coords[e].name] = new_result
    #Add equilibrium geometry and convert force units
    eq = gt.GLogInterpreter(log_files=[eq_log + '.log'])
    eq.pull_forces()
    forces = convert_force_units(eq_zmat=eq_zmat, force_constants=eq.force_constants)
    for r in results:
        results[r]['filenames'].append("eq")
        results[r]['disp_vals'].append(0)
        results[r]['disp_units'].append(results[r]['disp_units'][-1])
        results[r]['disp_energies'].append(0)
        results[r]['force_constants'] = [a * McData.UnitsData.convert("Hartrees", "Wavenumbers") for a in
                                         results[r]['force_constants']]
        if results[r]['disp_units'][0] ==  "Angstroms":
            results[r]['force_constants'] = [a * (1/McData.UnitsData.convert("BohrRadius", "Angstroms"))**2 for a in results[r]['force_constants']]
        elif results[r]['disp_units'][0] ==  "Degrees":
            results[r]['force_constants'] = [a * ((2 * np.pi) / 360)**2 for a in results[r]['force_constants']]
        for e in range(len(eq_zmat.coords)):
            if eq_zmat.coords[e].name == r:
                results[r]['force_constants'].append(forces[e][e])

    for key in results:
        zipped_lists = zip(results[key]['disp_vals'],
                           results[key]['filenames'],
                           results[key]['disp_units'],
                           results[key]['disp_energies'],
                           results[key]['force_constants'])
        sorted_pairs = sorted(zipped_lists) #sorts by items in the first list
        tuples = zip(*sorted_pairs)
        results[key]['disp_vals'], results[key]['filenames'],results[key]['disp_units'],results[key]['disp_energies'],results[key]['force_constants'] = [list(tuple) for tuple in tuples]

    # if sec_der:
    #     for key in results:
    #         results[key]['2nd_der'] = (results[key]['disp_energies'][0] + results[key]['disp_energies'][1])/results[key]['disp_vals'][0]**2
    return results



def convert_force_units(eq_zmat, force_constants):
    force_constants = force_constants * McData.UnitsData.convert("Hartrees", "Wavenumbers")
    for i in range(len(force_constants)):
        if eq_zmat.coords[i].units == "Angstroms":
            force_constants[i] = force_constants[i] * (1 / McData.UnitsData.convert("BohrRadius", "Angstroms"))
            force_constants[:, i] = force_constants[:, i] * (1 / McData.UnitsData.convert("BohrRadius", "Angstroms"))
        elif eq_zmat.coords[i].units == "Degrees":
            force_constants[i] = force_constants[i] * (2 * np.pi) / 360
            force_constants[:, i] = force_constants[:, i] * (2 * np.pi) / 360
    return force_constants

def target_displacements(eq_zmat, force_constants, disp_en = 20):
    #Gaussian force constants are in Hartree/Bohr^2 or Hartree/Radian^2
    force_constants = convert_force_units(eq_zmat = eq_zmat, force_constants=force_constants)
    disps = np.zeros(len(eq_zmat.coords))
    for d in range(len(disps)):
        disps[d] = np.sqrt(abs(2 * disp_en / force_constants[d][d]))
    return disps

def write_displaced_zmats(eq_zmat,
                          force_constants,
                          disp_en = 20,
                          sbatch_params='',
                          fname_base='',
                          params={},
                          job_file_path='',
                          fchks=False):
    with open(job_file_path + f'g16_{fname_base}_{disp_en}.sh', 'w', newline='\n') as sbatch:
        sbatch.write('#!/bin/bash\n')
        sbatch.write(sbatch_params)
        sbatch.write('\n# load Gaussian environment\nmodule load contrib/g16.b01\n\n# '
                     'debugging information\necho '
                     '\"**** Job Debugging Information ****\"\necho \"This job will '
                     'run on $SLURM_JOB_NODELIST"\necho \"\"\n'
                     'echo "ENVIRONMENT VARIABLES"\nset\necho \"********************'
                     '**************************\"\n\n# run Gaussian\n')
        disps = target_displacements(eq_zmat=eq_zmat, force_constants=force_constants, disp_en=disp_en)
        for d in range(len(disps)):
            pm = [disps[d], -disps[d]]
            for i in range(2):
                new_zmat = copy.deepcopy(eq_zmat)
                new_zmat.coords[d].val += pm[i]
                if len(str(new_zmat.coords[d].val)) > 8:
                    fname = fname_base + f"_{new_zmat.coords[d].name}_{str(new_zmat.coords[d].val).replace('.', 'p')[0:8]}"
                else:
                    fname = fname_base + f"_{new_zmat.coords[d].name}_{str(new_zmat.coords[d].val).replace('.', 'p')}"
                job_params = copy.deepcopy(params)
                job_params['header'] += f'{fname}.chk\n'
                job_params['molecule'] = new_zmat.print_mat()
                job_params['vars'] = new_zmat.print_vals()
                job = gt.GaussianJob(params=job_params, filename=f'{job_file_path}{fname}.gjf')
                job.save_job()
                sbatch.write(f'g16 {fname}.gjf\n')
                if fchks:
                    sbatch.write(f'formchk {fname}.chk\n')




class int_coord:
    def __init__(self, name=None, spec=None, val=None, units=None):
        self.name=name
        self.spec=spec
        self.val=val
        self.units=units


class zmat:

    #Example Build Dict
    # build_dict={"atoms" : ['H', 'O', 'N', 'O'],
    #             "specs" :  np.array([[1, 2, 3], [1, 2], [1]]),
    #             "names" : np.array([['r1', 'r2', 'r3'], ['a1', 'a2'], ['d3']]),
    #             "vals" : np.array([[0.97026274, 1.42776435, 1.17680217],
    #                                [101.62778541, 110.70859147],
    #                                [180.]]),
    #             "units" : ["Angstroms", "Degrees"]}

    def __init__(self, atoms=None, coords=None, charge=[0,1], easy_build_dict=None, g_str=None):
        self.atoms=atoms
        self.coords=coords
        self.charge=charge
        self.build_dict=easy_build_dict
        self.g_str = g_str
        if self.coords is None and self.build_dict is not None:
            self.coords=[]
            specs = []# build #N-6 array that specifies all connections
            for i in range(len(self.build_dict['specs'])):
                for j in range(len(self.build_dict['specs'][i])):
                    new_spec = []
                    new_spec.append(j+i+2)
                    for k in range(i+1):
                        new_spec.append(self.build_dict['specs'][k][j+i-k])
                    specs.append(new_spec)
            i=0
            u = 0
            for a in range(len(self.build_dict['names'])):
                for b in range(len(self.build_dict['names'][a])):
                    if len(specs[i]) > 2:
                        u = 1
                    else:
                        u = 0
                    self.coords.append(int_coord(name=self.build_dict['names'][a][b],
                                                 spec=specs[i],
                                                 val=self.build_dict['vals'][a][b],
                                                 units=self.build_dict['units'][u]))
                    i += 1
        elif self.coords is None and self.g_str is not None:
            self.coords = []
            self.atoms = []
            lines = g_str.splitlines()
            i = 0
            n = 0
            b = []
            a = []
            d = []

            for line in lines:
                if 'Charge' in line:
                    self.charge[0] = line.split()[2]
                    self.charge[1] = line.split()[5]
                    i = 1
                    continue
                elif 'Variables' in line:
                    i = 2
                    for bond in b:
                        self.coords.append(bond)
                    for ang in a:
                        self.coords.append(ang)
                    for dih in d:
                        self.coords.append(dih)
                    continue
                elif i == 1:
                    n += 1
                    l = line.split()
                    self.atoms.append(l[0])
                    if len(l) == 3:
                        b.append(int_coord(name=l[2], spec = [n, int(l[1])], units="Angstroms"))
                    elif len(l) == 5:
                        b.append(int_coord(name=l[2], spec=[n, int(l[1])], units="Angstroms"))
                        a.append(int_coord(name=l[4], spec=[n, int(l[1]), int(l[3])], units="Degrees"))
                    elif len(l) > 5:
                        b.append(int_coord(name=l[2], spec=[n, int(l[1])], units="Angstroms"))
                        a.append(int_coord(name=l[4], spec=[n, int(l[1]), int(l[3])], units="Degrees"))
                        d.append(int_coord(name=l[6], spec=[n, int(l[1]), int(l[3]), int(l[5])], units="Degrees"))
                elif i == 2:
                    l = line.split()
                    for c in self.coords:
                        if c.name == l[0]:
                            c.val = float(l[1])
        elif self.coords is None and self.g_str is None:
            print(f'self.coords is None and self.g_str is None')
            print(self.atoms)






    def print_mat(self):
        # z_mat = ''
        # z_mat += self.atoms[0] + '\n'
        # if len(self.atoms) > 1:
        #     z_mat += f"{self.atoms[1]} {self.coords[0].spec[-1]} {self.coords[0].name}" + '\n'
        # if len(self.atoms) > 2:
        #     z_mat += f"{self.atoms[2]} {self.coords[1].spec[-1]} {self.coords[1].name} {self.refs[2][0]} {self.vars[1][0]}" + '\n'
        # if len(self.atoms) > 3:
        #     for i in range(3, len(self.refs[0])):
        #         z_mat += f"{self.refs[0][i]} {self.refs[1][i - 1]} {self.vars[0][i - 1]} {self.refs[2][i - 2]} {self.vars[1][i - 2]} {self.refs[3][i - 3]} {self.vars[2][i - 3]}" + '\n'
        # return z_mat

        #assume order of coords is all bonds, all angles, all dihedrals as they appear on z-mat
        pre_mat = [[''],[''],['',''],['',''],['','',''],['','','']]
        for coord in self.coords:
            if len(coord.spec) == 2:
                pre_mat[0].append(coord.spec[-1])
                pre_mat[1].append(coord.name)
            elif len(coord.spec) == 3:
                pre_mat[2].append(coord.spec[-1])
                pre_mat[3].append(coord.name)
            elif len(coord.spec) == 4:
                pre_mat[4].append(coord.spec[-1])
                pre_mat[5].append(coord.name)

        mat = np.full((len(self.atoms),7), '', dtype='<U10') #weirdness with strings and data types. cuts off string if more than 10 chars long
        mat[:,0] = self.atoms
        for i in range(6):
            mat[:,i+1] = pre_mat[i]

        mat_str = ''
        for i in range(len(mat)):
            line=''
            for e in range(len(mat[i])):
                line += mat[i][e] + ' '
            mat_str += line + '\n'

        return mat_str

    def print_vals(self):
        vals = ''
        for coord in self.coords:
            vals += f' {coord.name}={coord.val}\n'
        return vals















